/*
 * rtc-ds13307.c - RTC driver for the DS1307 and DS1337 I2C chips.
 *
 * Copyright (C) 2018 Helmut Pozimski
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/module.h>
#include <linux/rtc.h>
#include <linux/i2c.h>
#include <linux/bcd.h>
#include <linux/slab.h>

#define M_NAME "rtc-ds13307" /* Module name */

#define DEVICE_DS1307 1
#define DEVICE_DS1337 2

#define COMMON_SEC 0x00
#define COMMON_MIN 0x01
#define COMMON_HOUR 0x02
#define COMMON_DAY 0x03
#define COMMON_DATE 0x04
#define COMMON_MONTH 0x05
#define COMMON_YEAR 0x06
#define COMMON_HIGH_BIT 0x80

#define DS1307_NVRAM_BASE 0x08
#define DS1307_MAX_ADDR 0x3F

#define DS1337_CTL 0x0E
#define DS1337_STAT 0x0F

static int model_detected;

static struct i2c_driver ds13307_driver;
static struct nvmem_config ds1307_nvmem;

/* Reads a specified number of bytes via i2c, returns 0 on success */
static int ds13307_read_bytes(struct i2c_client *client,
		u8 *addr, u8 *bytes, int length) {
	int r;
	struct i2c_msg msgs[] = {
		{
			.addr = client->addr,
			.len = 1,
			.buf = addr
		},
		{
			.addr = client->addr,
			.flags = I2C_M_RD,
			.len = length,
			.buf = bytes
		}
	};
	r = i2c_transfer(client->adapter, msgs, 2);
	if (r == 2) {
		return 0;
	} else {
		return r;
	}
}

/* Writes a specified number of bytes via 2ic, returns 0 on success */
static int ds13307_write_bytes(struct i2c_client *client,
	u8 *addr, u8 *bytes, int length) {
	int r;
	u8 *buf;
	struct i2c_msg msg;
	buf = (u8*) kmalloc(length +1, GFP_KERNEL);
	if (buf == NULL) {
		printk(KERN_ERR "%s: Could not allocate memory for buffer\n", M_NAME);
		return -EIO;
	}
	buf[0] = (*addr);
	memmove(buf + 1, bytes, length);
	msg.addr = client->addr;
	msg.len = length + 1;
	msg.buf = addr;
	r = i2c_transfer(client->adapter, &msg, 1);
	kfree(buf);
	if (r == 1) {
		return 0;
	} else {
		return r;
	}
}

/* The oscillator is stopped for both chips when power is first applied,
 * therefore this function checks its status and clears the stop bit.
 */
static int ds13307_start_oscillator(struct i2c_client *client) {
	u8 data, addr, buf;
	int r, v = 0;
	if (model_detected == DEVICE_DS1307) {
		addr = COMMON_SEC;
	} else {
		addr = DS1337_STAT;
	}
	r = ds13307_read_bytes(client, &addr, &data, 1);
	if (data & COMMON_HIGH_BIT) {
		buf = data ^ COMMON_HIGH_BIT;
		v = ds13307_write_bytes(client, &addr, &buf, 1);
		if (!v) {
			printk(KERN_DEBUG "%s: oscillator stop bit successfully cleared\n", M_NAME);
		}
	}
	if ((r<0) || (v)) {
		return -EIO;
	}
	return 0;
}

static int ds13307_read_time(struct device *dev, struct rtc_time *time) {
	struct i2c_client *client;
	int r, century = 1, h12 = 0;
	u8 buf[7], stopbit;
	u8 addr = COMMON_SEC;
	client = to_i2c_client(dev);
	r = ds13307_read_bytes(client, &addr, buf, 7);
	if (model_detected == DEVICE_DS1337) {
		addr = DS1337_STAT;
		r = ds13307_read_bytes(client, &addr, &stopbit, 1);
		stopbit = stopbit & COMMON_HIGH_BIT;
	} else {
		stopbit = buf[COMMON_SEC] & COMMON_HIGH_BIT;
	}
	if (r) {
		return -EIO;
	}
	if (stopbit) {
		printk(KERN_ERR "%s: Oscillator stop bit is set, values read from rtc device cannot be trusted\n", M_NAME);
		return -EINVAL;
	}
	if (buf[COMMON_HOUR] & 0x40) {
		buf[COMMON_HOUR] = buf[COMMON_HOUR] ^ 0x40;
		if (buf[COMMON_HOUR] & 0x20) {
			buf[COMMON_HOUR] = buf[COMMON_HOUR] ^ 0x20;
			h12 = 1;
		}
	}
	if (model_detected == DEVICE_DS1337) {
		century = buf[COMMON_MONTH] & COMMON_HIGH_BIT;
		buf[COMMON_MONTH] = buf[COMMON_MONTH] & 0x7F;
	}

	time->tm_sec = bcd2bin(buf[COMMON_SEC]);
	time->tm_min = bcd2bin(buf[COMMON_MIN]);
	time->tm_hour = bcd2bin(buf[COMMON_HOUR]);
	if(h12) {
		time->tm_hour += 12;
	}
	time->tm_mday = bcd2bin(buf[COMMON_DATE]);
	time->tm_wday = bcd2bin(buf[COMMON_DAY]);
	time->tm_mon = bcd2bin(buf[COMMON_MONTH]) - 1 ;
	if (century) {
		time->tm_year = bcd2bin(buf[COMMON_YEAR]) + 100;
	} else {
		time->tm_year = bcd2bin(buf[COMMON_YEAR]);
	}
	return 0;
}

static int ds13307_set_time(struct device *dev, struct rtc_time *time) {
	struct i2c_client *client;
	u8 buf[7], addr;
	client = to_i2c_client(dev);
	if (ds13307_start_oscillator(client)) {
		printk(KERN_ERR "%s: failed to initialize the oscillator\n", M_NAME);
		return -EIO;
	}
	addr = COMMON_SEC;
	buf[0] = bin2bcd(time->tm_sec);
	buf[1] = bin2bcd(time->tm_min);
	buf[2] = bin2bcd(time->tm_hour);
	buf[3] = bin2bcd(time->tm_wday);
	buf[4] = bin2bcd(time->tm_mday);
	if ((model_detected == DEVICE_DS1337) && (time->tm_year >= 100)) {
		buf[5] = bin2bcd(time->tm_mon + 1) | COMMON_HIGH_BIT;
	} else if ((model_detected == DEVICE_DS1307) && (time->tm_year < 100)) {
		       printk(KERN_ERR "%s: device does not support century information, dates before 2000 are not possible\n", M_NAME);
		       return -EINVAL;
	} else {
		buf[5] = bin2bcd(time->tm_mon + 1);
	}
	buf[6] = bin2bcd(time->tm_year % 100);
	if(ds13307_write_bytes(client, &addr, buf, 7)) {
		return -EIO;
	}
	return 0;
	
}

static const struct rtc_class_ops ds13307_rtc_ops = {
	.read_time = ds13307_read_time,
	.set_time = ds13307_set_time
};

/* Performs the device detection to distinguish between the
 * DS1307 and DS1337 chips. Returns the defined device ID or
 * -1 on failure
 */
static int ds13307_detect_device(struct i2c_client *client) {
	u8 data, addr = DS1307_MAX_ADDR;
	int result;
	result = ds13307_read_bytes(client, &addr, &data, 1);
	if (!result) {
		printk(KERN_INFO "%s: Detected device DS1307\n", M_NAME);
		return DEVICE_DS1307;
	}
        addr = DS1337_CTL;
	result = ds13307_read_bytes(client, &addr, &data, 1);
	if (!result) {	
		printk(KERN_INFO "%s: Detected device DS1337\n", M_NAME);
 		return DEVICE_DS1337;
	}
	printk(KERN_ERR "%s: Could not talk to I2C device at addr %x, is it connected?\n",
			M_NAME, client->addr);
	return -1;
}

static int ds13307_nvram_read(void *priv, unsigned int offset, void *buf, size_t count) {
	struct i2c_client *client = priv;
	u8 addr = DS1307_NVRAM_BASE + offset;
	if (!ds13307_read_bytes(client, &addr, buf, count)) {
		return count;
	}
	return -EIO;
}

static int ds13307_nvram_write(void *priv, unsigned int offset, void *buf, size_t count) {
	return 0;
}

static int ds13307_probe(struct i2c_client *client,
		const struct i2c_device_id *id) {
	struct rtc_device *rtc;
	int error;
	model_detected = ds13307_detect_device(client);
	if ((model_detected != DEVICE_DS1307) && (model_detected != DEVICE_DS1337)) {
		return -EIO;
	}
	rtc = devm_rtc_allocate_device(&client->dev);
	if (IS_ERR(rtc)) {
		return PTR_ERR(rtc);
	}
	if (model_detected == DEVICE_DS1307) {
		ds1307_nvmem.name = "ds1307_nvram";
		ds1307_nvmem.word_size = 1;
		ds1307_nvmem.size = 56;
		ds1307_nvmem.reg_read = ds13307_nvram_read;
		ds1307_nvmem.priv = client;
		rtc->nvmem_config = &ds1307_nvmem;
	}
	rtc->ops = &ds13307_rtc_ops;
	i2c_set_clientdata(client,rtc);
	error = rtc_register_device(rtc);
	return error ? error: 0;
}

static struct i2c_device_id ds13307_idtable[] = {
	{ "ds1307", 0 },
	{ "ds1337", 0 }, {}
};
MODULE_DEVICE_TABLE(i2c, ds13307_idtable);

static struct of_device_id ds13307_of_match[] = {
	{ .compatible = "dallas,ds1307" },
	{ .compatible = "dallas,ds1337" },
	{}
};
MODULE_DEVICE_TABLE(of, ds13307_of_match);

static struct i2c_driver ds13307_driver = {
	.driver = {
		.name = M_NAME,
		.of_match_table = of_match_ptr(ds13307_of_match),
		.owner = THIS_MODULE
	},
	.id_table = ds13307_idtable,
	.probe = ds13307_probe,
};

module_i2c_driver(ds13307_driver);

MODULE_AUTHOR("Helmut Pozimski <helmut@pozimski.eu>");
MODULE_DESCRIPTION("DS1307 and DS1337 RTC driver");
MODULE_LICENSE("GPL");
